import heap.Heap;

import java.util.Scanner;
import java.util.Map;
import java.util.HashMap;
import java.io.File;
import java.io.FileNotFoundException;


/** A program to demonstrate Huffman coding trees. */
public class Huff {

  // Problem 1: Count frequencies
  /** Counts the frequency of each character in a string. Returns a map from
   * single-character Strings to the number of times that character appears in
   * the input.*/
  public static HashMap<String,Integer> frequency(String input) {
    Scanner sc = new Scanner(input);
    HashMap<String,Integer> freq = new HashMap<String,Integer>();

    for (int i = 0; i < input.length(); i++) {
      String chr = input.substring(i,i+1);
      freq.put(chr, freq.getOrDefault(chr, 0) + 1);
    }
    return freq;
  }

  // Problem 2: Build tree
  /** Builds a Huffman coding tree using the given character frequencies.
   * Returns the root node of the Huffman tree. */
  public static HuffNode buildTree(HashMap<String,Integer> frequencies) {
    Heap<HuffNode,Integer> h = new Heap<HuffNode,Integer>();
    for (Map.Entry<String,Integer> entry : frequencies.entrySet()) {
      String word = entry.getKey();
      int freq = entry.getValue();
      HuffNode n = new HuffNode(null, null, freq, word);
      h.add(n, freq);
    }

    while (h.size() != 1) {
      HuffNode z = new HuffNode();
      z.left = h.poll();
      z.right = h.poll();
      z.freq = z.left.freq + z.right.freq;
      h.add(z, z.freq);
    }
    return h.poll();
  }

  // Problem 3: Decode
  /** Decode the encoded String code using the tree given by root */
  public static String decode(HuffNode root, String code) {
    StringBuilder result = new StringBuilder();
    HuffNode current = root;
    for (int i = 0; i < code.length(); i++) {
      char c = code.charAt(i);
      if (c == ' ') continue;
      if (c == '0') {
        current = current.left;
      } else {
        current = current.right;
      }
      if (current.word != null) {
        result.append(current.word);
        current = root;
      }
    }
    return result.toString();

  }

  // Problem 4: Build a dictionary.
  /** Traverses a coding tree, building a codec that maps words to codewords.
   *  prefix is the code representing the traversal from the global tree's root
   *  to the current node. */
  public static void getDictionary(HuffNode node, HashMap<String,String> codec, String prefix) {
    if (node.word != null) {
      codec.put(node.word, prefix);
      return;
    }
    getDictionary(node.left, codec, prefix + "0");
    getDictionary(node.right, codec, prefix + "1");
  }

  // Problem 5: Encode
  /** Encode the input using the given codec */
  public static String encode(HashMap<String,String> codec, String input) {
    StringBuilder result = new StringBuilder();
    for (int i = 0; i < input.length(); i++) {
      result.append(codec.get(input.substring(i,i+1)));
    }
    return result.toString();
  }



  /** Main method: reads text from a file, builds a huffman tree, encodes it, and
   * calculates compression ratio. */
  public static void main(String[] args) {

    // Create a Scanner to read the input file
    Scanner sc;
    try {
      File file = new File(args[0]);
      sc = new Scanner(file);
    } catch (FileNotFoundException e) {
      System.out.println("Could not find file");
      return;
    }

    // Read the input file into a single string:
    StringBuilder sb = new StringBuilder();
    while(sc.hasNextLine()) {
      String nl = sc.nextLine();
      if (sc.hasNextLine()) {
        sb.append(nl + System.lineSeparator());
      } else {
        sb.append(nl);
      }
    }
    String input = sb.toString();

    // Calculate frequencies:
    HashMap<String,Integer> frequencies = frequency(input);
    for (String k : frequencies.keySet()) {
      System.out.println(k + ": " + frequencies.get(k));
    }

    // Build a Huffman coding tree:
    HuffNode tree = buildTree(frequencies);
    tree.dump(0);

    // Build a map for easy lookup of the code for a given word:
    HashMap<String,String> codec = new HashMap<String,String>();
    getDictionary(tree, codec, "");

    for (String word : frequencies.keySet()) {
      System.out.println(word + ": " + codec.get(word));
    }

    // Encode the input string
    String coded = encode(codec, input);
    String output = decode(tree, coded);
    if (input.length() < 2000) {
      System.out.println("Input string: " + input);
      System.out.println("Encoded string: " + coded);
      System.out.println("Decoded string: " + output);
    }

    // Verify input matches output and print compression ratio
    System.out.println("input.equals(output): " + input.equals(output));

    // we divide by 8 because each input character takes 1 byte to store
    // and we could pack 8 output bits into a single byte
    double compression = (double)coded.length() / (double)input.length() / 8;
    System.out.println("Compression ratio: " + compression);
  }


  /** A class representing a node of a huffman coding tree.*/
  private static class HuffNode {
    HuffNode left; // child
    HuffNode right; // child
    int freq; // frequency count of all words in this subtree
    String word; // if this is a leaf, word is encoded as code, otherwise word is null

    /** Constructor: initialize to defaults */
    public HuffNode() {}

    /** Constructor: use this to create a leaf node */
    public HuffNode(HuffNode left, HuffNode right, int freq, String word) {
      this.left = left;
      this.right = right;
      this.freq = freq;
      this.word = word;
    }

    /** Print the tree sideways, indented by 4*level spaces */
    public void dump(int level) {
      if (left != null) {
        left.dump(level+1);
      }
      for (int i = 0; i < level; i++) {
        System.out.print("    ");
      }
      System.out.print(freq);
      if (word != null) {
        System.out.print(" " + word);
      }
      System.out.println();
      if (right != null) {
        right.dump(level+1);
      }
    }
  }

}
